/*
aPACK reimplementation
Written by and copyright (C) Joel Yliluoma in 2011; http://iki.fi/bisqwit/

This program reimplements the compression and decompression
algorithm used by Joergen Ibsen's aPACK version 1.0.
Specifically, its default algorithm...

With the following modification: End of compressed stream
is indicated the same way as in aPLib, rather than the way
it is done in aPACK.

*/
#include <algorithm>
#include <vector>
#include <cstdio>
#include <assert.h>
#include <cstring>

#include <unistd.h>
#include <stdlib.h>

#ifdef _OPENMP
# include <omp.h>
#endif

bool DoVerySlow = false;

namespace
{
    static size_t MatchingLength
        (const unsigned char* in1,
         size_t size,
         const unsigned char* in2)
    {
        // Trivial implementation would be:
        //   return std::mismatch(in1,in1+size, in2).second-in2;
        // However this function attempts to gain something
        // by comparing sizeof(unsigned) bytes at time instead
        // of 1 byte at time.
        size_t len = 0;
        typedef size_t T;
        const size_t intsize = sizeof(T);
        const T*const p1 = reinterpret_cast<const T*>(in1);
        const T*const p2 = reinterpret_cast<const T*>(in2);
        while(size >= intsize && p1[len] == p2[len]) { ++len; size -= intsize; }
        len *= intsize;
        while(size > 0 && in1[len] == in2[len]) { --size; ++len; }
        return len;
    }

    static unsigned GammaScore(size_t value)
    {
        assert(value >= 2);
        unsigned highest_bit = 0x80000000u;
        while(highest_bit != 0
           && !(value & highest_bit))
            highest_bit >>= 1;
        if(value & highest_bit)
            highest_bit >>= 1;
        unsigned num_bits_sent = 0;
        for(; ; ++num_bits_sent)
        {
            ++num_bits_sent;
            highest_bit >>= 1;
            if(!highest_bit) break;
        }
        return ++num_bits_sent;
    }

    static unsigned Cost(size_t offset, size_t length, size_t prev_offset, size_t default_cost)
    {
        if(length < 2) return default_cost;
        if(offset == prev_offset)
        {
            // Penalty of an append
            const unsigned GammaScore2  = 2;
            return 2 + GammaScore2 + GammaScore(length);
        }
        if(length >= 2 && length <= 3 && offset >= 1 && offset <= 127)
        {
            // Penalty of a short match
            return 3 + 8;
        }
        // Penalty of a long match
        if(offset < 0x80 && length < 4) return 100; // Penalize illegal combination

        size_t tmp = offset + 0x300, tmp2 = length;
        if(offset <= 127) tmp2 -= 2;
        return 2 + 8 + GammaScore(tmp >> 8) + GammaScore(tmp2);
    }
}

class Apack
{
public:
    std::vector<unsigned char> result;
    unsigned BitsRemaining, BitsPosition;
private:
    // First: copy length, second: begin offset
    std::vector< std::pair<size_t,size_t> > MaxMatchLen;
    std::vector< unsigned char >            DefaultCost;
    std::vector< std::pair<size_t,size_t> > Decisions;
    std::vector<bool> bomb; // bytes which must be encoded verbatim
                            // due to containing author's name

    std::vector<unsigned short> MatchCache;

private:
    size_t TestMatch(const std::vector<unsigned char>& source,
                     size_t pos1,
                     size_t pos2,
                     size_t limit)
    {
        if(pos2 > pos1) { size_t tmp = pos1; pos1=pos2; pos2=tmp; }
        const size_t p = pos1 * source.size() + pos2;
        if(MatchCache[p]) return MatchCache[p] - 1;

        size_t result = MatchingLength(&source[pos1], limit, &source[pos2]);
        MatchCache[ p ] = 1+result;
        return result;
    }

    void AnalyzeSource(const std::vector<unsigned char>& source,
                       unsigned constbyte)
    {
        const size_t endpos  = source.size();
        MaxMatchLen.resize(endpos);
        DefaultCost.resize(endpos);
        Decisions.resize(endpos);
        bomb.resize(endpos);

        MatchCache.resize(endpos * endpos);

        // Find out all prospects
        for(size_t a=1; a<endpos; ++a)
        {
            unsigned default_cost = 9;
            if(source[a] == constbyte)
                default_cost = 7;
            else
                for(size_t offset = 1; a >= offset && offset <= 15; ++offset)
                    if(source[a] == source[a-offset])
                        { default_cost = 7; break; }
            DefaultCost[a] = default_cost;

            size_t limit = endpos - a;
            size_t longest = 1, where = 0;
            for(size_t test=0; test<a; ++test)
            {
                size_t matchlen = TestMatch(source, test,a, limit);
                if(matchlen > longest)
                {
                    longest = matchlen;
                    where   = test;
                }
            }
            MaxMatchLen[a].first  = longest;
            MaxMatchLen[a].second = where;
        }
        static const char* const MagicWords[] =
        {
            "bisqwit"/*, "Joel Yliluoma"*/, "Joel", "Yliluoma"
        };
        for(unsigned tokenno=0; tokenno < sizeof(MagicWords)/sizeof(*MagicWords); ++tokenno)
        {
            const char* const token = MagicWords[tokenno];
            const size_t token_size = strlen(token);

            size_t begin = 0;
            while(begin + token_size <= source.size())
            {
                void* p = memmem( &source[begin], source.size()-begin,
                                  token, token_size);
                if(!p) break;
                size_t found_pos = (const unsigned char*)p - &source[0];
                for(unsigned a=0; a<token_size; ++a)
                {
                    MaxMatchLen[found_pos + a].first = 1;
                    bomb[found_pos + a] = true;
                    DefaultCost[found_pos + a] = 9;
                }
                begin = found_pos + token_size;
            }
        }

        // Update match-cache for bombs
        for(size_t a=1; a<endpos; ++a)
        {
            size_t limit = endpos-a;
            for(size_t test=0; test<a; ++test)
            {
                size_t len = TestMatch(source, a,test, limit);
                for(size_t p=0; p<len; ++p)
                {
                    if(bomb[test+p])
                    {
                        MatchCache[a * endpos + test] = p;
                        break;
                    }
                }
            }
        }
    }

    unsigned DetermineDecision(const std::vector<unsigned char>& source,
                               size_t a = 1,
                               size_t test_last = ~size_t(0),
                               unsigned recursion_limit = 2)
    {
        const size_t endpos  = source.size();

        unsigned total_cost = 0;

        // Determine decision
        size_t make_fast_when = std::min((a+endpos) / 2, a + 256);
        unsigned n_actions = 0;
        while(a < endpos)
        {
            if(recursion_limit == 2 && (a >= make_fast_when || ++n_actions >= 2))
                recursion_limit -= 1;
            if(recursion_limit == 1 && (a >= make_fast_when || ++n_actions >= 20))
                recursion_limit -= 1;

            unsigned best_cost   = 0;
            unsigned my_cost     = 0;
            size_t   best_eat    = MaxMatchLen[a].first;
            size_t   best_source = MaxMatchLen[a].second;

            if(!recursion_limit)
            {
                // Simply take predefine maximum (be greedy)
                my_cost = Cost(a - MaxMatchLen[a].second,
                               MaxMatchLen[a].first,
                               test_last,
                               DefaultCost[a]);
            }
            else
            {
              #ifdef _OPENMP
                omp_lock_t lock;
                omp_init_lock(&lock);
              #endif
               #pragma omp parallel for \
                    schedule(guided) \
                    default(none) \
                    shared(best_cost,best_eat,best_source,my_cost,source,lock,DoVerySlow) \
                    firstprivate(a,recursion_limit,test_last)
                for(size_t eating_count = best_eat; eating_count >= 1; --eating_count)
                {
                    if(eating_count < 2)
                    {
                        size_t cheapest_source = MaxMatchLen[a].second;
                        unsigned cheapest_cost = DefaultCost[a];

                        unsigned trailing_cost = cheapest_cost + DetermineDecision(
                            source,
                            a + eating_count,
                            eating_count >= 2 ? a-cheapest_source : test_last,
                            recursion_limit - 1);

                      #ifdef _OPENMP
                        omp_set_lock(&lock);
                       #pragma omp flush(best_cost,best_eat,best_source,my_cost)
                      #endif
                        if(!best_cost || trailing_cost < best_cost)
                        {
                            best_cost   = trailing_cost;
                            best_eat    = eating_count;
                            best_source = cheapest_source;
                            my_cost     = cheapest_cost;
                        }
                      #ifdef _OPENMP
                       #pragma omp flush(best_cost,best_eat,best_source,my_cost)
                        omp_unset_lock(&lock);
                      #endif
                    }
                    else if(DoVerySlow)
                    {
                        for(size_t test=0; test<a; ++test)
                        {
                            if(TestMatch(source, test,a, eating_count) >= eating_count)
                            {
                                /*for(size_t p=0; p<eating_count; ++p)
                                    if(bomb[test+p]) goto discard_match;*/

                                size_t cheapest_source = test;
                                unsigned cheapest_cost = Cost(a-test, eating_count, test_last, DefaultCost[a]);

                                unsigned trailing_cost = cheapest_cost + DetermineDecision(
                                    source,
                                    a + eating_count,
                                    eating_count >= 2 ? a-cheapest_source : test_last,
                                    recursion_limit - 1);

                              #ifdef _OPENMP
                                omp_set_lock(&lock);
                               #pragma omp flush(best_cost,best_eat,best_source,my_cost)
                              #endif
                                if(!best_cost || trailing_cost < best_cost)
                                {
                                    best_cost   = trailing_cost;
                                    best_eat    = eating_count;
                                    best_source = cheapest_source;
                                    my_cost     = cheapest_cost;
                                }
                              #ifdef _OPENMP
                               #pragma omp flush(best_cost,best_eat,best_source,my_cost)
                                omp_unset_lock(&lock);
                              #endif
                            }
                        }
                    }
                    else
                    {
                        size_t cheapest_source = 0;
                        unsigned cheapest_cost = 0;
                        for(size_t test=0; test<a; ++test)
                            if(TestMatch(source, test,a, eating_count) >= eating_count)
                            {
                                /*for(size_t p=0; p<eating_count; ++p)
                                    if(bomb[test+p]) goto discard_match;*/

                                unsigned cost = Cost(a-test, eating_count, test_last, DefaultCost[a]);
                                if(!cheapest_cost || cost < cheapest_cost)
                                {
                                    cheapest_cost   = cost;
                                    cheapest_source = test;
                                }
                            }

                        unsigned trailing_cost = cheapest_cost + DetermineDecision(
                            source,
                            a + eating_count,
                            eating_count >= 2 ? a-cheapest_source : test_last,
                            recursion_limit - 1);

                      #ifdef _OPENMP
                        omp_set_lock(&lock);
                       #pragma omp flush(best_cost,best_eat,best_source,my_cost)
                      #endif
                        if(!best_cost || trailing_cost < best_cost)
                        {
                            best_cost   = trailing_cost;
                            best_eat    = eating_count;
                            best_source = cheapest_source;
                            my_cost     = cheapest_cost;
                        }
                      #ifdef _OPENMP
                       #pragma omp flush(best_cost,best_eat,best_source,my_cost)
                        omp_unset_lock(&lock);
                      #endif
                    }
                }
              #ifdef _OPENMP
                omp_destroy_lock(&lock);
              #endif
            }

            Decisions[a].first  = best_eat;
            Decisions[a].second = best_source;
            if(best_eat >= 2) test_last = a - best_source;
            a          += best_eat;
            total_cost += my_cost;
        }
        return total_cost;
    }
public:
    void Compress(const std::vector<unsigned char>& source, unsigned max_recursion,
                  unsigned constbyte)
    {
        const size_t endpos  = source.size();

        AnalyzeSource(source, constbyte);

        DetermineDecision(source, 1, ~size_t(0), max_recursion);

        // Encode the decision
        result.clear();
        BitsRemaining = 0;
        BitsPosition  = 0;
        size_t readpos = 0;
        size_t last_offset = 0;
        result.push_back( source[readpos++] );

        while(readpos < endpos)
        {
            const size_t copy_length = Decisions[readpos].first;
            const size_t copy_begin  = Decisions[readpos].second;

            if(copy_length < 2)
            {
                unsigned char byte = source[readpos];
                /*fprintf(stderr, "[%u]: literal %02X\n", (unsigned)readpos, byte);*/
                if(byte == constbyte)
                    SendByteCopy(0);
                else
                {
                    if(!bomb[readpos])
                    {
                        for(size_t offset = 1; readpos >= offset && offset <= 15; ++offset)
                            if(byte == source[readpos-offset])
                                { SendByteCopy(offset); goto didbytecopy; } // Range: 1..15
                    }
                    SendLiteral(byte);
                didbytecopy:;
                }
                ++readpos;
                continue;
            }

            size_t my_offset = readpos - copy_begin;

            // Append is *always* better than LongMatch. (6 vs 11 bits)
            // ShortMatch is *always* better than LongMatch. (11 vs 12+ bits)
            if(my_offset == last_offset)
            {
                // Length: 2..N
                SendAppendMatch( copy_length ); // Range: 1..inf
            }
            else if(copy_length >= 2 && copy_length <= 3
                 && my_offset   >= 1 && my_offset <= 127)
            {
                // Range: 1..127, 2..3
                SendShortMatch( my_offset, copy_length );
            }
            else
            {
                // Length: M..N
                // where M is 1 for offset=80..4FF
                //            2 for offset=500..7CFF
                //            3 for others
                if(my_offset < 0x80) assert(copy_length >= 4);
                else                 assert(copy_length >= 2);
                assert(copy_length < 0x100);
                assert(my_offset   >= 1);
                SendLongMatch( my_offset, copy_length );
            }
            last_offset = my_offset;
            readpos += copy_length;
        }
        SendEnd();
    }
    void Uncompress(const std::vector<unsigned char>& source,
                    unsigned n_calls, unsigned transformation_mask,
                    unsigned constbyte)
    {
        unsigned tagpos=0, tag=0;
        size_t inpos = 0;

        #define GetBit() \
           ( ( (tagpos%8 ? (tag                  ) << (tagpos++ % 8) \
                         : (tag = source[inpos++]) << (tagpos++ % 8) \
               ) & 0x80) >> 7)

        result.clear();
        size_t last_offs = 0;
    Literal:
        result.push_back(source[inpos++]);
        for(;;)
        {
            if(! GetBit() ) goto Literal; // 0, literal
            if(! GetBit() ) // 10, codepair
            {
                unsigned bh = 1; do bh = bh*2 + GetBit(); while( GetBit() );
                unsigned cx = 1; do cx = cx*2 + GetBit(); while( GetBit() );
                if(bh >= 3) // offshi=2, append; otherwise, offslo
                {
                    unsigned bx = ((bh-3) << 8) | source[inpos++];
                    if(bx <= 0x7F) cx += 2;
                    last_offs = bx;
                }
                do result.push_back( result[result.size() - last_offs] );
                while(--cx > 0);
            }
            else if(! GetBit() ) // 110, bytecopy
            {
                unsigned bx = 0;
                bx |= GetBit() << 3;
                bx |= GetBit() << 2;
                bx |= GetBit() << 1;
                bx |= GetBit() << 0;
                if(!bx)
                    result.push_back(constbyte);
                else
                    result.push_back( result[result.size() - bx] );
            }
            else // 111, shortmatch
            {
                unsigned bx = source[inpos++];
                unsigned cx = 3 - (bx & 1);
                last_offs = bx >> 1;
                if(!last_offs) break;
                do result.push_back( result[result.size() - last_offs] );
                while(--cx > 0);
            }
        }
        for(size_t a=0; a+3<=result.size() && n_calls > 0; ++a)
        {
            if(((transformation_mask & 1) && result[a] == 0xE8)
            || ((transformation_mask & 2) && result[a] == 0xE9))
            {
                ++a;
                unsigned offset = result[a] | (result[a+1] << 8);
                offset -= (a+0x100);
                result[a]   = offset & 0xFF;
                result[a+1] = offset >> 8;
                --n_calls;
                ++a;
            }
        }
        if(n_calls)
            fprintf(stderr, "%u calls were not found\n", n_calls);
    }
private:
    // Primitives:
    void PrimSendByte(unsigned char value)
    {
        result.push_back(value);
    }
    void PrimSendBit(unsigned char bit)
    {
        if(!BitsRemaining)
            { BitsPosition = result.size();
              PrimSendByte(0x00);
              BitsRemaining = 8; }
        result[BitsPosition] |= bit << --BitsRemaining;
        /*fprintf(stderr, "sending bit: %02X -- tag@%u is now %02X\n", bit, BitsPosition,result[BitsPosition]);*/
    }
    void PrimSendGamma(unsigned value) // The smallest value it can encode is 2.
    {
        assert(value >= 2);
        unsigned highest_bit = 0x80000000u;
        while(highest_bit != 0
           && !(value & highest_bit))
            highest_bit >>= 1;
        if(value & highest_bit)
            highest_bit >>= 1;
        for(; ; PrimSendBit(1))
        {
            PrimSendBit( !! ( value & highest_bit) );
            highest_bit >>= 1;
            if(!highest_bit) break;
        }
        PrimSendBit(0);
    }
public:
    // Actual encodes
    void SendLongMatch(
        unsigned offset,
            // Offset: 0..n
        unsigned len
            // Len, if offset:
            //    0..7F:    3..any (1 + 2)
            //   80..FFFF:  1..any (1 + 0)
    )
    {
        /*fprintf(stderr, "Sending long match: %u %u\n", offset, len);*/
        PrimSendBit(1); PrimSendBit(0);
        unsigned tmp = offset + 0x300;
        PrimSendGamma(tmp >> 8);
        if(offset <= 127) len -= 2;
        PrimSendGamma(len);
        PrimSendByte(tmp & 0xFF);
    }
    void SendAppendMatch(unsigned len) // Only valid after SendLiteral or SendByteCopy.
    {
        /*fprintf(stderr, "Sending APPEND: %u\n", len);*/
        PrimSendBit(1);  PrimSendBit(0);
        PrimSendGamma(2);
        PrimSendGamma(len);
    }
    void SendShortMatch(unsigned offset/*1..127*/, unsigned len/*2..3*/)
    {
        /*fprintf(stderr, "Sending short match: %u %u\n", offset, len);*/
        PrimSendBit(1); PrimSendBit(1); PrimSendBit(1);
        PrimSendByte( (3-len) | (offset << 1) );
    }
    void SendByteCopy(unsigned offset/*1..15*/)
    {
        /*fprintf(stderr, "Sending byte copy: %u\n", offset);*/
        PrimSendBit(1); PrimSendBit(1); PrimSendBit(0);
        PrimSendBit( (offset >> 3) & 1 );
        PrimSendBit( (offset >> 2) & 1 );
        PrimSendBit( (offset >> 1) & 1 );
        PrimSendBit( (offset >> 0) & 1 );
    }
    void SendLiteral(unsigned char literal)
    {
        /*fprintf(stderr, "Sending literal: %02X\n", literal);*/
        PrimSendBit(0);
        PrimSendByte(literal);
    }
    void SendEnd()
    {
        //PrimSendBit(1); PrimSendBit(0); PrimSendGamma(0x100);
        // ^ 16 bits
        SendShortMatch(0, 2); // 11 bits
    }
};


/*
    Patterns:
        %1$c   ';' if no call patching is used, ' ' otherwise
        %2$c   ' ' if constvalue = 0, ';' otherwise
        %3$c   ';' if constvalue = 0, ' ' otherwise
        %4$d   value of constvalue
        %5$s   'cx' if ncalls >= 0x100, 'cl' otherwise
        %6$d   value of ncalls
        %7$d   0xE9 if transformation_mask == 2, 0xE8 otherwise
        %8$d   ' ' if transformation_mask == 3, ';' otherwise
*/

static const char StubTemplate[] =
"bits 16\n"
"STUB_LENGTH equ 0xD ; (end_stub-begin_stub)\n"
"org 0x7F80 - STUB_LENGTH ; 7F80*2 = FF00; -FF00 = 100. This is why 7F80.\n"
"    begin_stub:\n"
"	mov si,0x100 + STUB_LENGTH\n"
"	mov di,DECOMPRESS\n"
"	mov cx,di\n"
"	cld\n"
"	push di\n"
"	rep movsb\n"
"	ret\n"
"    end_stub:\n"
"DECOMPRESS:\n"
"	neg di ;mov di,0x100\n"
"%1$c	push di\n"
"	push di\n"
"	mov si, DATABEGIN\n"
"	mov dl,0x80\n"
"	mov bp, GETBIT\n"
"	push ax\n"
"LITERAL:\n"
"	movsb\n"
"NEXTTAG:\n"
"	call bp\n"
"	jnc LITERAL\n"
"	call bp\n"
"	inc cx\n"
"	jnc CODEPAIR\n"
"	call bp\n"
"	mov bx,0x10\n"
"	jc SHORTMATCH\n"
".fourbits:\n"
"	call bp\n"
"	adc bl,bl\n"
"	jnc .fourbits\n"
"	jnz DOMATCH_BX\n"
"%2$c	xchg ax,bx\n"
"%3$c	mov al, 0x%4$X\n"
"	jmp STORE\n"
"CODEPAIR:\n"
"	mov bh,1\n"
".bhloop:\n"
"	call bp\n"
"	adc bh,bh\n"
"	call bp\n"
"	jc .bhloop\n"
".cxloop:\n"
"	call bp\n"
"	adc cx,cx\n"
"	call bp\n"
"	jc .cxloop\n"
"	sub bh,3\n"
"	jc DOMATCH_LASTPOS\n"
"SHORTMATCH:\n"
"	mov bl,[si]\n"
"	inc si\n"
"	jnc .waslong\n"
"	shr bl,1\n"
"	jz OVER\n"
"	sbb cl,bh\n"
".waslong:\n"
"	cmp bx,0x7F\n"
"	ja DOMATCH_NEWLASTPOS\n"
"	inc cx\n"
"	inc cx\n"
"DOMATCH_NEWLASTPOS:\n"
"	pop ax\n"
"	db 0xA8 ; test al\n"
"DOMATCH_LASTPOS:\n"
"	pop bx\n"
"	push bx\n"
"DOMATCH_BX:\n"
"	neg bx\n"
"COPYLOOP:\n"
"	mov al,[bx+di]\n"
"STORE:\n"
"	stosb\n"
"	loop COPYLOOP\n"
"	jmp NEXTTAG\n"
"GETBIT:\n"
"	add dl,dl\n"
"	jnz .getbitret\n"
"	lodsb\n"
"	xchg ax,dx\n"
"	adc dl,dl\n"
".getbitret:\n"
"	ret\n"
"OVER:\n"
"	pop ax\n"
"%1$c 	pop si\n"
"%1$c .findcalls:\n"
"%1$c 	mov %5$s, %6$d\n"
"%1$c .findcall:\n"
"%1$c 	lodsb\n"
"%1$c %8$c and al,0xFE\n"
"%1$c 	cmp al, 0x%7$X\n"
"%1$c 	jnz .findcall\n"
"%1$c 	sub [si],si\n"
"%1$c 	lodsw\n"
"%1$c 	loop .findcall\n"
"	ret\n"
"DATABEGIN:\n"
;

int main()
{
    std::vector<unsigned char> buffer_orig;
    for(;;)
    {
        size_t end = buffer_orig.size();
        buffer_orig.resize(end + 4096);
        size_t n = std::fread(&buffer_orig[end], 1, 4096, stdin);
        buffer_orig.resize(end + n);
        if(n == 0) break;
    }

    DoVerySlow = buffer_orig.size() < 2000;

    std::vector<unsigned char> result, buffer_calls;
    unsigned n_calls_translated = 0;
    unsigned transformation_mask = 0;
    unsigned n_calls_found = 0;
    unsigned n_jumps_found = 0;

    for(size_t a=0; a+3<=buffer_orig.size(); ++a)
    {
        if(buffer_orig[a] == 0xE8)
        {
            ++n_calls_found;
            a += 2;
        }
        else if(buffer_orig[a] == 0xE9)
        {
            ++n_jumps_found;
            a += 2;
        }
    }
  #ifdef _OPENMP
    omp_set_nested(0);
  #endif
  #pragma omp parallel for
    for(unsigned transformations=0; transformations<4; ++transformations)
    {
        std::vector<unsigned char> buffer(buffer_orig);
        for(size_t a=0; a+3<=buffer.size(); ++a)
        {
            if(((transformations & 1) && buffer[a] == 0xE8)
            || ((transformations & 2) && buffer[a] == 0xE9))
            {
                ++a;
                unsigned offset = buffer[a] | (buffer[a+1] << 8);
                offset += (a+0x100);
                buffer[a]   = offset & 0xFF;
                buffer[a+1] = offset >> 8;
                ++a;
            }
        }
        Apack packer;
        packer.Compress(buffer, 0, 0x00);
        std::vector<unsigned char>& test_result = packer.result;

      #pragma omp critical
      {
        unsigned my_calls = 0;
        if(transformations & 1) my_calls += n_calls_found;
        if(transformations & 2) my_calls += n_jumps_found;
        size_t res_size  = result.size();
        size_t test_size = test_result.size();
        if(my_calls > 0) test_size += 14;
        if(my_calls > 0xFF) ++test_size;
        if(transformations == 3) test_size += 2;
        if(n_calls_translated > 0)    res_size += 14;
        if(n_calls_translated > 0xFF) ++res_size;
        if(transformation_mask == 3) test_size += 2;

        if(result.empty() || test_size < res_size)
        {
            buffer_calls.swap(buffer);
            result.swap(test_result);
            n_calls_translated = my_calls;
            transformation_mask = transformations;
        }
        #pragma omp flush(n_calls_translated)
      }
    }
    //fprintf(stderr, "Translated %u calls\n", n_calls_translated);
    unsigned constbyte = 0x00;

 #pragma omp parallel for
    for(unsigned test_constbyte = 0x00; test_constbyte < 0x100; ++test_constbyte)
    {
        Apack packer;
        packer.Compress(buffer_calls, 0, test_constbyte);
        std::vector<unsigned char>& test_result = packer.result;
      #pragma omp critical
      {
        size_t res_size  = result.size();
        size_t test_size = test_result.size();
        if(constbyte != 0x00) ++res_size;
        if(test_constbyte != 0x00) ++test_size;
        if(result.empty() || test_size < res_size)
        {
            result.swap(packer.result);
            constbyte = test_constbyte;
        }
        #pragma omp flush(constbyte)
      }
    }

    Apack packer;
    packer.Compress(buffer_calls, 3, constbyte);
    result.swap(packer.result);

#if 1
    Apack unpacker;
    unpacker.Uncompress(result, n_calls_translated,transformation_mask, constbyte);

    std::vector<unsigned char> decompress( unpacker.result);

    bool ok = true;
    {
        unsigned r = decompress.size();

        if(buffer_orig.size() != r) ok = false;
        fprintf(stderr, "Input was %u bytes, compressed to %u bytes <- %u bytes. %u calls,calljump mask %u, constbyte=%02X\n",
            (unsigned) buffer_orig.size(),
            (unsigned) result.size(),
            r,
            n_calls_translated,transformation_mask, constbyte);

        unsigned nmismatch=0, firstfail=~0u, lastfail=~0u;
        for(unsigned a=0; a<r; ++a)
            if(buffer_orig[a] != decompress[a])
            {
                ++nmismatch;
                if(firstfail==~0u) firstfail=a;
                lastfail=a;
            }
        if(nmismatch)
        {
            ok = false;
            fprintf(stderr, "%u failures, within %u..%u range (total size: %u)\n",
                nmismatch,firstfail,lastfail, r);
        }
    }
    if(!ok) fprintf(stderr, "= ERROR\n");
#else
    /*
    fprintf(stderr, "Input was %u bytes, compressed to %u bytes\n",
        (unsigned) buffer_orig.size(),
        (unsigned) result.size());*/
#endif

    char tempfn1[64], tempfn2[64];
    sprintf(tempfn1, "apack-%d.tm1", getpid());
    sprintf(tempfn2, "apack-%d.tm2", getpid());

    FILE* fp = fopen(tempfn1, "wt");
    fprintf(fp,
        StubTemplate,
        n_calls_translated == 0 ? ';' : ' ',
        constbyte == 0x00 ? ' ' : ';',
        constbyte == 0x00 ? ';' : ' ',
        constbyte,
        n_calls_translated >= 0x100 ? "cx" : "cl",
        n_calls_translated,
        transformation_mask == 2 ? 0xE9 : 0xE8,
        transformation_mask == 3 ? ' ' : ';'
    );
    fclose(fp);

    char command[512];
    sprintf(command, "nasm -w+all -O2 %1$s -o %2$s && cat %2$s; rm -f %1$s %2$s", tempfn1,tempfn2);
    system(command);

    std::fwrite(&result[0], 1, result.size(), stdout);
    return 0;
}
